/**
 * 	@file	SM130.cpp
 * 	@brief	SM130 library for Arduino
 *	@author	Fabio Angeletti <fabio.angeletti89@gmail.com>
 *	@date	November 2012
 *
 *	<p>
 *	Controls a SonMicro SM130
 *	</p>
 *
 *	@see	http://www.arduino.cc
 *	@see	http://www.sonmicro.com
 */

#include "SM130.h"

/** Constructor
 * 
 */
SM130::SM130(){
	_Serial = NULL;
}

///////////////////////////////////
///		begin communication		///
///////////////////////////////////
void SM130::begin(HardwareSerial *serIn){
	_Serial = serIn;
	_Serial->begin(DEFAULT_BAUD_RATE);
}

///////////////////////////////////////////////////////
/// function to flush unwanted responses from SM130 ///
///////////////////////////////////////////////////////
void SM130::cleanSerial(){
	while(_Serial->available()){
		_Serial->read();
		delay(5);
	}
}

//////////////////////////////////////////
/// reset for SM130                    ///
/// returns:                           ///
///     0   if unknown error occurred  ///
///     1   if reset is successful     ///
///    -1   if checksum error          ///
//////////////////////////////////////////
int SM130::reset(){
	cleanSerial();
	sendCommand(resetID);
	
	getData();

	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}
	
	if(dataBuffer[3] == 0x81){
		return 1;
	}
	return 0;
}

//////////////////////////////////////////
/// get SM130 firmware version         ///
/// returns:                           ///
///     0   if unknown error occurred  ///
///     1   if command is successful   ///
///    -1   if checksum error          ///
//////////////////////////////////////////
int SM130::firmware(){  
	cleanSerial();
	sendCommand(firmwareReadID);
	
	getData();

	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}
	
	if(dataBuffer[3] == 0x81){
		return 1;
	}
	return 0;
}

/////////////////////////////////////////////
/// halt TAG                              ///
/// returns:                              ///
///     0   if unknown error occurred     ///
///     1   if PICC is halted successful  ///
///    -1   if checksum error             ///
/////////////////////////////////////////////
int SM130::haltTAG(){
	cleanSerial();
	sendCommand(haltID);
	
	getData();

	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}
	
	if(dataBuffer[3] == 0x93){
		return 1;
	}
	return 0;
}

//////////////////////////////////////////////
/// select TAG                             ///
///    and fill TAG serial and type        ///
/// returns:                               ///
///     0   if unknown error occurred      ///
///     1   if command executed correctly  ///
///    -1   if checksum error              ///
///    -2   if no TAG is present           ///
///    -3   if RF field is OFF             ///
//////////////////////////////////////////////
int SM130::selectTAG(){
	int i = 0;
	cleanSerial();

	currentTAGType = 0x00;

	sendCommand(selectTagID);

	getData();

	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}

	if(dataBuffer[2] == 0x02 && dataBuffer[3] == 0x83){
		switch (dataBuffer[4]){
			case 0x4E:
			return -2;
			break;
			case 0x55:
			return -3;
			break;
			default:
			return 0;
			break;
		}
	}

	currentTAGType = dataBuffer[4];
	switch(dataBuffer[4]){
		case mifare1k:
		case mifare4k:
		for(i;i<4;i++)
			currentTAGSerial[i] = dataBuffer[5+i];
		break;
		case mifareUltralight:
		for(i;i<7;i++)
			currentTAGSerial[i] = dataBuffer[5+i];
		default:
		break;
	}

	for(i;i<7;i++)
		currentTAGSerial[i] = 0;
	
	return 1;
}

////////////////////////////////////
/// seek for TAG                 ///
////////////////////////////////////
void SM130::seekForTAG(){
	cleanSerial();
	sendCommand(seekForTagID);
}

//////////////////////////////////////////
/// SM130 enter sleep mode             ///
/// resurrect only on hw reset         ///
/// returns:                           ///
///     1   if entered sleep state     ///
///     0   if unknown error occurred  ///
///    -1   if checksum error          ///
//////////////////////////////////////////
int SM130::enterSleep(){
	cleanSerial();
	sendCommand(sleepID);
	
	getData();

	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}
	
	if(dataBuffer[3] == 0x96){
		return 1;
	}
	return 0;
}

/////////////////////////////////////////
/// switch antenna power on SM130     ///
/// returns:                          ///
///     2  if RF field is ON          ///
///     1  if RF field is OFF         ///
///     0  if unknown error occurred  ///
///    -1   if checksum error         ///
/////////////////////////////////////////
int SM130::switchAntenna(byte aStatus){
	cleanSerial();
	sendCommand(antennaPowerSwitchID, aStatus);
	
	getData();

	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}
	
	if(dataBuffer[3] == 0x90){
		if(dataBuffer[4] == 0x00){
			return 1;
		}
		if(dataBuffer[4] >= 0x01){
			return 2;
		}
	}
	return 0;
}

////////////////////////////////////////////
/// read input port                      ///
/// returns:                             ///
///     int  port status                 ///
///    -1    if checksum error           ///
///    -2    if unknown error occurred   ///
////////////////////////////////////////////
int SM130::readPort(){
	cleanSerial();
	sendCommand(readFromPortID);
	
	getData();
	
	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}
	
	if(dataBuffer[3] == 0x91){
		return dataBuffer[4];
	}
	return -2;
}

////////////////////////////////////////////
/// write output port                    ///
/// returns:                             ///
///     int   current port status        ///
///    -1     if checksum error          ///
///    -2     if unknown error occurred  ///
////////////////////////////////////////////
int SM130::writePort(byte aStatus){
	cleanSerial();
	sendCommand(writeToPortID, aStatus);
	
	getData();
	
	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}

	if(dataBuffer[3] == 0x92){
		return dataBuffer[4];
	}
	return -2;
}

//////////////////////////////////////////////
/// function to check packet integrity     ///
/// return:                                ///
///     1    if packet is valid            ///
///     0    if packet is invalid          ///
///    -1    if no valid packet is found   ///
//////////////////////////////////////////////
int SM130::checkIntegrity(byte *array, int startPos){
	byte checkSum = 0x00;
	int index = startPos;
	int payloadLength = 0;

	//find packet's start
	while((array[index] != 0xFF) || (array[index+1] != 0x00) || (index < (MAX_BUFFER_SIZE-2))){
		index++;
		if(index >= (MAX_BUFFER_SIZE-2)){
		//no valid packet found
			return -1;
		}
	}

	payloadLength = array[index+2];
	for(int i=(index+3);i<(index+3+payloadLength);i++){
		checkSum += array[i];
	}
	if(checkSum == array[index+3+payloadLength+1]){
		//checksum correct
		return 1;
	}else{
		//checksum incorrect
		return 0;   
	} 
}

//////////////////////////////////////////
/// receive from SM130                 ///
/// returns:                           ///
///    int    how many bytes received  ///
//////////////////////////////////////////
int SM130::getData(){
	int counter = 0x00;

	while(_Serial->available() <= 0);
	while(_Serial->available() > 0){
		dataBuffer[counter] = _Serial->read();
		counter++;
		delay(3);
	}
	return counter;
}

///////////////////////////////////////////////////
/// authenticate        TAG                     ///
/// returns:                                    ///
///     0    if unknown error occurred          ///
///     1    if authentication is successful    ///
///    -1    if checksum error                  ///
///    -2    if No TAG present or login failed  ///
///    -3    if login failed                    ///
///    -4    if invalid key format in EEPROM    ///
///////////////////////////////////////////////////
int SM130::authenticate(byte blockNum, byte keyType, byte *keyArray){
	byte payload[8];

	cleanSerial();

	payload[0] = blockNum;
	payload[1] = keyType;
	for(int i=0;i<6;i++){
		payload[2+i] = keyArray[i];
	}  
	if (keyType == 0xAA || keyType == 0xBB){
		sendCommand(authenticateID, payload, 8);
	}else{
		sendCommand(authenticateID, payload, 2);
	}
	
	getData();
	
	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}
	
	if(dataBuffer[2] = 0x02 && dataBuffer[3] == authenticateID){
		switch (dataBuffer[4]){
			case 0x4C:
			return 1;
			break;
			case 0x4E:
			return -2;
			break;
			case 0x55:
			return -3;
			break;
			case 0x45:
			return -4;
			break;
			default:
			return 0;
			break;
		}
	}
	return 0;
}

///////////////////////////////////////////////////
/// authenticate        TAG                     ///
/// returns:                                    ///
///     0    if unknown error occurred          ///
///     1    if authentication is successful    ///
///    -1    if checksum error                  ///
///    -2    if No TAG present or login failed  ///
///    -3    if login failed                    ///
///    -4    if invalid key format in EEPROM    ///
///////////////////////////////////////////////////
int SM130::authenticate(byte blockNum){
	byte payload[2];

	cleanSerial();

	payload[0] = blockNum;
	payload[1] = keyTypeF;
	sendCommand(authenticateID, payload, 2);
	
	getData();
	
	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}
	
	if(dataBuffer[2] = 0x02 && dataBuffer[3] == authenticateID){
		switch (dataBuffer[4]){
			case 0x4C:
			return 1;
			break;
			case 0x4E:
			return -2;
			break;
			case 0x55:
			return -3;
			break;
			case 0x45:
			return -4;
			break;
			default:
			return 0;
			break;
		}
	}
	return 0;
}

//////////////////////////////////////////
/// read block                         ///
/// returns:                           ///
///     1   if read is successful      ///
///     0   if unknown error occurred  ///
///    -1   if checksum is invalid     ///
///    -2   if no TAG is present       ///
///    -3   if read is failed          ///
//////////////////////////////////////////
int SM130::readBlock(byte blockNum){
	cleanSerial();
	sendCommand(readSpecifiedBlockID, blockNum);
	
	getData();
	
	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}
	
	if(dataBuffer[2] == 0x12){
		return 1;
	}
	
	if(dataBuffer[2] == 0x02){
		switch (dataBuffer[3]){
			case 0x4E:
			return -2;
			break;
			case 0x46:
			return -3;
			break;
			default:
			return 0;
			break;
		}
	}
	return 0;
}

//////////////////////////////////////////////
/// write block                            ///
/// returns:                               ///
///     1	if write is successful         ///
///     0   if unknown error occurred      ///
///    -1   if checksum error              ///
///    -2   if read after write failed     ///
///    -3   if unable to read after write  ///
///    -4   if no TAG is present           ///
///    -5   if write is failed             ///
//////////////////////////////////////////////
int SM130::writeBlock(byte blockNum, byte *dataArray){
	byte arrayToSend[BLOCK_SIZE+1];
	
	cleanSerial();
	
	arrayToSend[0] = blockNum;
	for(int i=0;i<BLOCK_SIZE;i++){
		arrayToSend[i+1] = dataArray[i];
	}
	sendCommand(writeDataBlockID, arrayToSend, BLOCK_SIZE+1);
	
	getData();
	
	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}

	if(dataBuffer[2] == 0x12){
		return 1;
	}else{
		switch(dataBuffer[4]){
			case 0x55:
			return -2;
			break;
			case 0x58:
			return -3;
			break;
			case 0x4E:
			return -4;
			break;
			case 0x46:
			return -5;
			break;
			default:
			return 0;
			break;      
		}
	}
	return 0;
}

//////////////////////////////////////////////
/// write four bytes block                 ///
/// returns:                               ///
///     1	if write is successful         ///
///     0   if unknown error occurred      ///
///    -1   if checksum error              ///
///    -2   if read after write failed     ///
///    -3   if unable to read after write  ///
///    -4   if no TAG is present           ///
///    -5   if write is failed             ///
//////////////////////////////////////////////
int SM130::writeFourByteBlock(byte blockNum, byte *dataArray){
	byte arrayToSend[5];
	
	cleanSerial();
	
	arrayToSend[0] = blockNum;
	for(int i=0;i<4;i++){
		arrayToSend[i+1] = dataArray[i];
	}
	sendCommand(writeDataBlockID, arrayToSend, 5);
	
	getData();
	
	if(!checkIntegrity(dataBuffer, 0)){
		return -1;
	}

	if(dataBuffer[2] == 0x12){
		return 1;
	}else{
		switch(dataBuffer[4]){
			case 0x55:
			return -2;
			break;
			case 0x58:
			return -3;
			break;
			case 0x4E:
			return -4;
			break;
			case 0x46:
			return -5;
			break;
			default:
			return 0;
			break;      
		}
	}
	return 0;
}

///////////////////////////////////
// return null terminated string //
// indicating the TAG type       //
///////////////////////////////////
const char* SM130::getTAGType(){
	switch(currentTAGType){
		case mifareUltralight:
			return "Mifare Ultralight";
			break;
		case mifare1k:
			return "Mifare Standard 1k";
			break;
		case mifare4k:
			return "Mifare Classic 4k";
		case tagTypeUnknown:
		default:
			return "Unknown TAG";
			break;
	}
}

//////////////////////////////////////////
//check if a block is the trailer block //
//////////////////////////////////////////
int SM130::isTrailerBlock(int blockNum){
	switch(currentTAGType){
		case mifare1k:
		if(blockNum % 4 == 3)
			return 1;
		break;
		case mifare4k:
		if(blockNum <= 128){
			if(blockNum % 4 == 3)
				return 1;
		}else{
			if(blockNum % 16 == 15)
				return 1;
		}
		break;
		default:
		break;
	} 
	return 0;
}

////////////////////////////////////
/// sendCommand with OVERLOADING ///
///       optional serial debug  ///
////////////////////////////////////
void SM130::sendCommand(byte cmd, byte *param, int paramLen){
	int checkSum = 0x00;

	cleanSerial();
	
	pktData[0] = 0xFF;
	pktData[1] = 0x00;
	pktData[2] = paramLen + 1;
	pktData[3] = cmd;

	checkSum += paramLen + 1;
	checkSum += cmd;

	if(paramLen > 0){
		for(int j=0; j<paramLen; j++){
			pktData[4+j] = param[j];
			checkSum += param[j];
		}
	}
	pktData[4+paramLen] = checkSum;

	for(int k=0; k<(paramLen+5); k++){
		_Serial->write(pktData[k]);
	}
}

void SM130::sendCommand(byte cmd, byte param){
	int checkSum = 0x00;

	cleanSerial();
	
	pktData[0] = 0xFF;
	pktData[1] = 0x00;
	pktData[2] = 0x02;
	pktData[3] = cmd;

	checkSum += 2;
	checkSum += cmd;

	pktData[4] = param;
	checkSum += param;
	pktData[5] = checkSum;

	for(int k=0; k<6; k++){
		_Serial->write(pktData[k]);
	}	
}

void SM130::sendCommand(byte cmd){
	int checkSum = 0x00;

	cleanSerial();
	
	pktData[0] = 0xFF;
	pktData[1] = 0x00;
	pktData[2] = 0x01;
	pktData[3] = cmd;

	checkSum += 1;
	checkSum += cmd;

	pktData[4] = checkSum;

	for(int k=0; k<5; k++){
		_Serial->write(pktData[k]);
	}
}

