/**
 * 	@file	SM130.h
 * 	@brief	Header file for SM130 library
 *	@author	Fabio Angeletti <fabio.angeletti89@gmail.com>
 *	@date	November 2012
 */

#ifndef SM130_h
#define SM130_h

#if defined(ARDUINO) && ARDUINO >= 100
#include "Arduino.h"
#else
#include "WProgram.h"
#endif

//general purpose buffer size
#define MAX_BUFFER_SIZE                         64

//block parameters
#define BLOCK_SIZE                              16

//serial communication default baud rate
#define DEFAULT_BAUD_RATE						19200

//complete list of commands from SM130 datasheet
 enum sm130_command_t{
	antennaPowerSwitchID = 0x90,			//Antenna Power 		Switches ON or OFF the RF field
	authenticateID = 0x85, 					//Authenticate 			Authenticates the selected Block
	baudrateID = 0x94,						//Set Baud Rate 		Sets the new baud rate
	decValueBlockID = 0x8E,					//Decrement 			Decrements a value block
	firmwareReadID = 0x81, 					//Firmware 				Reads the Firmware Revision of the Module
	haltID = 0x93,							//Halt 					Halts the PICC
	incValueBlockID = 0x8D,					//Increment 			Increments a value block
	NA1_ID = 0x84, 							//NA 					Not Implemented
	NA2_ID = 0x88, 							//NA 					Not Implemented
	NA3_ID = 0x8F,							//NA 					Not Implemented
	NA4_ID = 0x95,							//NA 					Not Implemented
	readFromPortID = 0x91,					//Read port 			Reads from the Input port
	readSpecifiedBlockID = 0x86,			//Read Block 			Reads from the specified Block
	readValueBlockID = 0x87,				//Read Value 			Reads from a Value Block
	resetID = 0x80, 						//Reset 				Resets the Module
	seekForTagID = 0x82, 					//Seek for Tag 			Continuously checks for presence of a tag
	selectTagID = 0x83, 					//Select Tag 			Selects a Tag
	sleepID = 0x96,							//Sleep 				This command puts SM130 in sleep mode
	writeChipEEPROMID = 0x8C,				//Write Key 			Writes the Key to the EEPROM of the MFRC530
	writeDataBlockID = 0x89, 				//Write Block 			Writes the data to the specified block
	writeFourBytesID = 0x8B,				//Write 4 Byte Block 	Writes 4 byte data to Mifare Ultralight block
	writeToPortID = 0x92,					//Write Port 			Writes to the Output port
	writeValueBlockID = 0x8A, 				//Write Value 			Formats and Writes a Value block
};

//possible types of TAG
enum tag_type_t {
  mifareUltralight = 0x01,					//Mifare Ultralight
  mifare1k = 0x02,							//Mifare Standard 1k
  mifare4k = 0x03,							//Mifare Classic 4k
  tagTypeUnknown = 0x04,					//Unknown tag type
};

//reply codes from various functions
enum status_code_t {
	statusSuccess = 0x00,
	statusInvalidResponse = 0x01,
	statusInvalidKey = 0x45,
	statusInProgress = 0x4C,
	statusLoginSuccessful = 0x4C,
	statusHalted = 0x4C,
	statusNoTag = 0x4E,
	statusRFOff = 0x55,
	statusLoginFailed = 0x55,
};

//complete list of supported keys
enum key_type_t {
	keyTypeA = 0xAA,
	keyTypeB = 0xBB,
	keyTypeF = 0xFF,
	keyTypeAStored = 0x10,
	keyTypeBStored = 0x20,
};

class SM130{
		
public:
		//current TAG information
	byte currentTAGType;
	byte currentTAGSerial[7];
		//global variables for buffering and packet operations
	byte dataBuffer[MAX_BUFFER_SIZE];
	byte pktData[MAX_BUFFER_SIZE];

		//! Constructor
	SM130();
		//! Initialize serial port
	void begin(HardwareSerial *serIn);
		//! Flush unwanted data in serial buffer
	void cleanSerial();
		//! Reset SM130 module
	int reset();
		//! Halt TAG chip
	int haltTAG();
		//! Select TAG
	int selectTAG();
		//! Seek for TAG
	void seekForTAG();
		//! Enter SM130 sleep mode
	int enterSleep();
		//! Switch ON/OFF RF field
	int switchAntenna(byte aStatus);
		//! Read input port
	int readPort();
		//! Write output port
	int writePort(byte aStatus);
		//! Receive data from SM130
	int getData();
		//! Authenticate block
	int authenticate(byte blockNum, byte keyType, byte *keyArray);
		//! Authenticate block with default transport key
	int authenticate(byte blockNum);
		//! Read block
	int readBlock(byte blockNum);
		//! Write block
	int writeBlock(byte blockNum, byte *dataArray);
		//! Write 4 byte block
	int writeFourByteBlock(byte blockNum, byte *dataArray);
		//! TAG type as null terminated string
	const char* getTAGType();
		//! TAG Serial as a pointer
	byte* getTAGSerial() { return currentTAGSerial; };
		//! Checks if a block is a trailer block
	int isTrailerBlock(int blockNum);
	

private:		
	HardwareSerial *_Serial;

		//! Get the module firmware
	int firmware();
		//! Check packet integrity
	int checkIntegrity(byte *array, int startPos);
		//! Send command
	void sendCommand(byte cmd, byte *param, int paramLen);
	void sendCommand(byte cmd, byte param);
	void sendCommand(byte cmd);
};

#endif